window.Gleam.initSearch([{"doc":"boyer_moore","title":"boyer_moore","content":"# boyer_moore\n\n[![Package Version](https://img.shields.io/hexpm/v/boyer_moore)](https://hex.pm/packages/boyer_moore)\n[![Hex Docs](https://img.shields.io/badge/hex-docs-ffaff3)](https://hexdocs.pm/boyer_moore/)\n\nAn implementation of the Boyer-Moore search algorithm for BitArrays and user-defined data structures.\n\nCurrently the algorithm only implements &quot;bad character&quot; rule.\n\n```sh\ngleam add boyer_moore\n```\n\n```gleam\nimport boyer_moore\nimport gleam/io\n\npub fn main() {\n  let needle = &lt;&lt;3, 4, 5&gt;&gt;\n\n  // Compile needle to searcher that can be reused\n  let assert Ok(searcher) = boyer_moore.build_byte_searcher(needle)\n\n  // Find needle in given BitArray\n  io.debug(boyer_moore.search(&lt;&lt;1, 2, 3, 4, 5, 6&gt;&gt;, searcher))\n  // Ok(2)\n}\n```\n\n## Performance for BitArray searching\n\nA naive version of searching through a BitArray in Gleam could be written like this:\n\n```gleam\nfn naive(data: BitArray, pattern: BitArray) {\n  do_naive(data, pattern, bit_array.byte_size(pattern), 0)\n}\n\nfn do_naive(data: BitArray, pattern: BitArray, pattern_size: Int, index: Int) {\n  case data {\n    &lt;&lt;&gt;&gt; -&gt; Error(Nil)\n    &lt;&lt;maybe_pattern:bytes-size(pattern_size), _rest:bytes&gt;&gt;\n      if maybe_pattern == pattern\n    -&gt; Ok(index)\n    &lt;&lt;_byte:size(8), rest:bytes&gt;&gt; -&gt;\n      do_naive(rest, pattern, pattern_size, index + 1)\n    _ -&gt; panic\n  }\n}\n```\n\nSummarised, the function iterates the BitArray byte by byte, and each time checks if the currently viewed section of\nthe BitArray starts with the pattern. This algorithm is what the sections below compare to.\n\n### Erlang\n\nOn the Erlang target, the BitArray API of this package delegates to\n[`binary:match/3`](https://www.erlang.org/doc/apps/stdlib/binary.html#match/3) and as such is much faster and more\nmemory efficient than the naive search on all haystack/needle sizes.\n\n### JavaScript\n\nIn the JavaScript target, the naive algorithm is actually faster for small inputs, meaning small haystacks and small\nneedles. The longer the needle, the more opportunities Boyer-Moore has for skipping forward. And the larger the\nhaystack, the more work the naive algorithm has to do.\n\nMeasured on an Apple Silicon M1 Max laptop with Node 20, the tipping point seems to be around the point where\n`haystack * needle` (as bytes) is 10000. For example, with a 26 byte needle and a 420 byte haystack, Boyer-Moore was\nabout 5 % faster than the naive algorithm. See the table below for more details:\n\n| Haystack length | Needle length | Relative performance |\n| --------------- | ------------- | -------------------- |\n| 55 B            | 6 B           | 57 % (i.e. slower)   |\n| 420 B           | 26 B          | 105 %                |\n| 1024 B          | 19 B          | 188 %                |\n| 30374 B         | 39 B          | 252 %                |\n\nAs a summary, for very short values, using the naive algorithm may be faster. On the other hand, in those cases the\nabsolute time difference will also be small. You may wish to benchmark how this algorithm works for your use case.\n\n## Usage for custom data structures\n\nYou can also use this package for searching inside a custom data structure. A suitable data structure offers fast\naccess to random indices regardless of the amount of items. As an example a Dict would work, but a List would not.\n\nAdditionally, the data structure _must_ have an integer index that runs from 0 to the amount of the items minus one,\nincrementing by one for each item and containing no holes. A given index must also always point to the same item\n(until the data structure is modified). This must stay true both for the needle and the haystack.\n\nTo use a custom data structure, you&#39;ll need to define an API that the library can use:\n\n```gleam\nimport boyer_moore\n\nconst dict_api = boyer_moore.API(\n  get_size: dict.size,\n  are_equal: boyer_moore.default_equality,\n  unsafe_access: dict_unsafe_get,\n)\n\nfn dict_unsafe_get(data: Dict(a, b), index: a) -&gt; b {\n  let assert Ok(item) = dict.get(data, index)\n  item\n}\n```\n\nNow you can create a searcher by passing in the API as well:\n\n```gleam\ntype DNA {\n  A\n  C\n  G\n  T\n}\n\nlet searcher = boyer_moore.build_generic_searcher(\n  dict.from_list([\n    #(0, A), #(1, A), #(2, G), #(3, T)\n  ]),\n  dict_api\n)\n\nboyer_moore.search(some_data, searcher)\n```\n","url":"index.html"},{"doc":"boyer_moore","title":"API","content":"pub type API(container, item) {\n  API(\n    get_size: fn(container) -&gt; Int,\n    are_equal: fn(item, item) -&gt; Bool,\n    unsafe_access: fn(container, Int) -&gt; item,\n  )\n}\n An API to use for a generic search over a user defined container type.\nAPI(\n  get_size: fn(container) -&gt; Int,\n  are_equal: fn(item, item) -&gt; Bool,\n  unsafe_access: fn(container, Int) -&gt; item,\n)\n\nget_size\n&lt;p&gt;Returns the amount of distinct items in the container.&lt;/p&gt;\n\nare_equal\n&lt;p&gt;Returns True if two items are equal to each other. You should use\n&lt;code&gt;default_equality&lt;/code&gt; – which calls Gleam’s equality operator – unless\nyou know that it won’t work for your item data type.&lt;/p&gt;\n\nunsafe_access\n&lt;p&gt;Returns item with the given index in the container. The algorithm will\nmake sure to only call this within the bounds of the container, so the\nfunction can return a value directly instead of a result (using\n&lt;code&gt;let assert&lt;/code&gt;). This assumes that the container does not have holes in\nit.&lt;/p&gt;\n\nSynonyms:\nboyer_moore.API\nboyer_moore API","url":"boyer_moore.html#API"},{"doc":"boyer_moore","title":"Bounds","content":"pub type Bounds {\n  Bounds(start: option.Option(Int), end: option.Option(Int))\n}\n Bounds for the search operation. If either value is `None` or out of range,\n they are set to the corresponding end index (i.e. `start` will be 0 and\n `end` will be the size of the data). If `end` is lower than `start`, it will\n be set to `start` (and thus the range to search will be empty).\n\n Negative indices are not supported.\nBounds(start: option.Option(Int), end: option.Option(Int))\n\n\nSynonyms:\nboyer_moore.Bounds\nboyer_moore Bounds","url":"boyer_moore.html#Bounds"},{"doc":"boyer_moore","title":"Searcher","content":"pub opaque type Searcher(container)\n A compiled searcher that can be reused many times on different data.\n\nSynonyms:\nboyer_moore.Searcher\nboyer_moore Searcher","url":"boyer_moore.html#Searcher"},{"doc":"boyer_moore","title":"build_byte_searcher","content":"pub fn build_byte_searcher(\n  pattern: BitArray,\n) -&gt; Result(Searcher(BitArray), Nil)\n Build a BitArray searcher that works for byte-aligned BitArrays.\n\n On Erlang, this delegates to `binary:match` as it is faster.\n\n Returns an error if the BitArray is empty, or -- on Erlang -- if the\n BitArray is not byte-aligned.\nSynonyms:\nboyer_moore.build_byte_searcher\nboyer_moore build_byte_searcher","url":"boyer_moore.html#build_byte_searcher"},{"doc":"boyer_moore","title":"build_generic_searcher","content":"pub fn build_generic_searcher(\n  pattern pattern: a,\n  api api: API(a, b),\n) -&gt; Result(Searcher(a), Nil)\n Build a generic searcher that can be used many times.\n\n This function allows you to use the search algorithm with your own custom\n data structures by implementing the functions in the `API` type.\n\n A suitable data structure is one that provides fast, preferably constant\n time access to random indices. For example, a Dict would work, a List would\n not.\n\n Additionally, the provided data structure must be 0-indexed with a\n strictly increasing index with no empty spaces (holes). If this guarantee is\n broken, the algorithm may crash.\n\n Returns an error if the pattern is empty.\nSynonyms:\nboyer_moore.build_generic_searcher\nboyer_moore build_generic_searcher","url":"boyer_moore.html#build_generic_searcher"},{"doc":"boyer_moore","title":"default_equality","content":"pub fn default_equality(a: a, b: a) -&gt; Bool\n Default equality algorithm that uses Gleam&#39;s equality operator (`==`). This\n will work for the majority of cases and should be used unless you know your\n data structure will not work with it.\nSynonyms:\nboyer_moore.default_equality\nboyer_moore default_equality","url":"boyer_moore.html#default_equality"},{"doc":"boyer_moore","title":"search","content":"pub fn search(\n  haystack data: a,\n  needle searcher: Searcher(a),\n) -&gt; Result(Int, Nil)\n Search for `needle` in the `haystack`. Returns the index to the start of the\n first occurrence of the found pattern, or an error if nothing was found.\nSynonyms:\nboyer_moore.search\nboyer_moore search","url":"boyer_moore.html#search"},{"doc":"boyer_moore","title":"search_bounded","content":"pub fn search_bounded(\n  haystack data: a,\n  needle searcher: Searcher(a),\n  bounds bounds: Bounds,\n) -&gt; Result(Int, Nil)\n Search for `needle` in the `haystack`, using `bounds` to limit the searched\n area. Returns the index to the start of the first occurrence of the found\n pattern, or an error if nothing was found.\nSynonyms:\nboyer_moore.search_bounded\nboyer_moore search_bounded","url":"boyer_moore.html#search_bounded"},{"doc":"boyer_moore","title":"boyer_moore","content":"","url":"boyer_moore.html"}]);